#include <M5Stack.h>
#include <Wire.h>
#include <Adafruit_MCP23X17.h>

//MCP23017 (GPIOエキスパンダ)の使い方
// http://project59.blog.fc2.com/blog-entry-441.html
// ポイント : ResetはVcc、アドレスはGNDに接続する事
#define L_H_JOY  7
#define L_V_JOY  6
#define R_H_JOY  5
#define R_V_JOY  4
#define A_BTN    8
#define B_BTN    9
#define X_BTN   10
#define Y_BTN   11
#define R_BTN   12
#define RZ_BTN  13 
#define L_BTN   14
#define LZ_BTN  15 
#define INPUT1_BTN   0
#define INPUT2_BTN   1
#define INPUT3_BTN   2
#define INPUT4_BTN   3

#define MAX_MODE_NUM 3          //キーコンフィグ設定数
#define RENDA_JUDGE_TIME 250    //連打と判定する長押し時間(ms)
#define RENDA_KANKAKU_MS 60     //連打する際のボタン押し間隔(ms)

Adafruit_MCP23X17 mcp;

uint32_t g_start_time;
uint32_t g_btnoff_time;   //ボタンが最後にOFFになっていた時間
bool g_thanks_flg=0;
uint32_t  g_wifi_interval_time = 0;

String cmd_msg;
String recv_buffer;

// jyuji_val : 十字ボタン [3] 上　[2] 下  [1]左 [0]右
// btn_val : ボタン　[7]LZ [6]L [5] RZ [4]R　[3] Y　[2] x [1]B [0]A
int jyuji_val, btn_val = 0;
int input_val, input_val_old = 0;
int input_cnt = 0;
int game_idx = 0;
int game_idx_old = -1;

void setup()
{
  M5.begin();
  Serial.begin(115200);
  M5.Lcd.fillScreen(BLACK);
  
  mcp.begin_I2C();

  //PIN設定
  mcp.pinMode(L_V_JOY, INPUT_PULLUP);
  mcp.pinMode(L_H_JOY, INPUT_PULLUP);
  mcp.pinMode(R_V_JOY, INPUT_PULLUP);
  mcp.pinMode(R_H_JOY, INPUT_PULLUP);
  
  mcp.pinMode(A_BTN, OUTPUT);
  mcp.digitalWrite(A_BTN, 1);
  
  mcp.pinMode(B_BTN, OUTPUT);
  mcp.digitalWrite(B_BTN, 1);
  
  mcp.pinMode(X_BTN, OUTPUT);
  mcp.digitalWrite(X_BTN, 1);
  
  mcp.pinMode(Y_BTN, OUTPUT);
  mcp.digitalWrite(Y_BTN, 1);
  
  mcp.pinMode(L_BTN, OUTPUT);
  mcp.digitalWrite(L_BTN, 1);
  
  mcp.pinMode(LZ_BTN, OUTPUT);
  mcp.digitalWrite(LZ_BTN, 1);
  
  mcp.pinMode(R_BTN, OUTPUT);
  mcp.digitalWrite(R_BTN, 1);
  
  mcp.pinMode(RZ_BTN, OUTPUT);
  mcp.digitalWrite(RZ_BTN, 1);

   // 時間リセット
   g_start_time = millis();  //時間リセット
   g_btnoff_time = millis();  //ボタンOFF検出時間のリセット
}

void loop(){
  
   M5.update();
   if(M5.BtnA.wasPressed()){  
      //現時点では3モードのみ
      game_idx = (game_idx-1)%MAX_MODE_NUM;     
   } 
   if(M5.BtnC.wasPressed()){  
      //現時点では3モードのみ
      game_idx = (game_idx+1)%MAX_MODE_NUM;     
   }
   
  //ゲームモード(キーコンフィグ)選択
  if(game_idx != game_idx_old){
    /////////////////////////////
    //モード0 : マリオカート8 DX
    ////////////////////////////
    if(game_idx == 0){
      M5.Lcd.drawJpgFile(SD, "/mariocart.jpg");
      M5.Lcd.drawJpgFile(SD, "/mc_accel.jpg", 0, 180);
      M5.Lcd.drawJpgFile(SD, "/mc_item.jpg", 80, 180);
      M5.Lcd.drawJpgFile(SD, "/mc_left.jpg", 160, 180);
      M5.Lcd.drawJpgFile(SD, "/mc_right.jpg", 240, 180);
    /////////////////////////////
    //モード1 : ゼルダの伝説
    /////////////////////////////
    }else if(game_idx == 1){
      M5.Lcd.drawJpgFile(SD, "/zelda.jpg");
      M5.Lcd.drawJpgFile(SD, "/zelda_a.jpg", 0, 180);
      M5.Lcd.drawJpgFile(SD, "/zelda_b.jpg", 80, 180);
      M5.Lcd.drawJpgFile(SD, "/zelda_x.jpg", 160, 180);
      M5.Lcd.drawJpgFile(SD, "/zelda_y.jpg", 240, 180);           
    }
    /////////////////////////////
    //モード2: 太鼓の達人
    /////////////////////////////
    else if(game_idx == 2){
      M5.Lcd.drawJpgFile(SD, "/taiko.jpg");
      M5.Lcd.drawJpgFile(SD, "/taiko_don.jpg", 0, 180);
      M5.Lcd.drawJpgFile(SD, "/taiko_don.jpg", 80, 180);
      M5.Lcd.drawJpgFile(SD, "/taiko_kaxtsu.jpg", 160, 180);
      M5.Lcd.drawJpgFile(SD, "/taiko_kaxtsu.jpg", 240, 180);           
    }
    game_idx_old = game_idx;
  }

  /////////////////////////////////
  //外部ボタンからのハック分などをマージ
  // → jyuji_val, btn_valに論理和でマージする
  // INPUT1 : Aボタン
  // INPUT2 : Bボタン
  // INPUT3 : 左ボタン
  // INPUT4 : 右ボタン

  int jyuji_val_update  = 0;
  int btn_val_update    = 0;
  int input_val = mcp.digitalRead(INPUT4_BTN)*8+mcp.digitalRead(INPUT3_BTN)*4+mcp.digitalRead(INPUT2_BTN)*2+mcp.digitalRead(INPUT1_BTN);


  /////////////////////////////
  //モード0 : マリオカート8 DX
  //  Input1: Aボタン(アクセル)
  //  Input2: Lボタン(アイテム)
  //  Input3: 左ボタン
  //  Input4: 右ボタン
  ////////////////////////////
  if(game_idx == 0){
    // ボタン入力
    if(mcp.digitalRead(INPUT1_BTN) == 0){
      btn_val_update |= 0b00000001; //Aボタンを1にする
    }
    if(mcp.digitalRead(INPUT2_BTN) == 0){
      btn_val_update |= 0b01000000; //Lボタンを1にする 
      //もしAが押されている場合は、下ボタンを押す = 後ろにアイテムを投げる
      if(btn_val_update & 0b00000001){
         jyuji_val_update |= 0b0100; //下ボタンを1にする    
      }
    }
    if(mcp.digitalRead(INPUT3_BTN) == 0){
      jyuji_val_update |= 0b0010; //左ボタンを1にする    
    }
    if(mcp.digitalRead(INPUT4_BTN) == 0){
      jyuji_val_update |= 0b0001; //右ボタンを1にする   
    }

    // ボタン表示変更
    if((input_val & 0b0001) != (input_val_old & 0b0001)){
      if((input_val & 0b0001) == 0) M5.Lcd.drawJpgFile(SD, "/mc_accel_on.jpg", 0, 180);
      else M5.Lcd.drawJpgFile(SD, "/mc_accel.jpg", 0, 180);
    }
    if((input_val & 0b0010)  != (input_val_old & 0b0010)){
      if((input_val & 0b0010) == 0) M5.Lcd.drawJpgFile(SD, "/mc_item_on.jpg", 80, 180);
      else M5.Lcd.drawJpgFile(SD, "/mc_item.jpg", 80, 180);
    }
    if((input_val & 0b0100) != (input_val_old & 0b0100)){
      if((input_val & 0b0100) == 0) M5.Lcd.drawJpgFile(SD, "/mc_left_on.jpg", 160, 180);
      else M5.Lcd.drawJpgFile(SD, "/mc_left.jpg", 160, 180);
    }
    if((input_val & 0b1000) != (input_val_old & 0b1000)){
      if((input_val & 0b1000) == 0) M5.Lcd.drawJpgFile(SD, "/mc_right_on.jpg", 240, 180);
      else M5.Lcd.drawJpgFile(SD, "/mc_right.jpg", 240, 180);
    }
    
  /////////////////////////////
  //モード1 : ゼルダの伝説
  //  Input1: Aボタン
  //  Input2: Bボタン
  //  Input3: Xボタン
  //  Input4: Yボタン
  ////////////////////////////
  }else if(game_idx == 1){
     if(mcp.digitalRead(INPUT1_BTN) == 0){
      btn_val_update |= 0b00000001; //Aボタンを1にする
    }
    if(mcp.digitalRead(INPUT2_BTN) == 0){
      btn_val_update |= 0b00000010; //Bボタンを1にする  
    }
    if(mcp.digitalRead(INPUT3_BTN) == 0){
      btn_val_update |= 0b00000100; //Xボタンを1にする    
    }
    if(mcp.digitalRead(INPUT4_BTN) == 0){
      btn_val_update |= 0b00001000; //Yボタンを1にする  
    }   
    
  /////////////////////////////
  //モード2 : 太鼓の達人
  //  Input1: 下ボタン(ドン)
  //  Input2: Yボタン(ドン)
  //  Input3: Lボタン(カッ)
  //  Input4: Rボタン(カッ)
  ////////////////////////////
  }else if(game_idx == 2){
    if(mcp.digitalRead(INPUT1_BTN) == 0){
       jyuji_val_update |= 0b0100; //下ボタンを1にする
    }
    if(mcp.digitalRead(INPUT2_BTN) == 0){
      btn_val_update |= 0b00001000; //Yボタンを1にする   
    }
    if(mcp.digitalRead(INPUT3_BTN) == 0){
      btn_val_update |= 0b01000000; //Lボタンを1にする    
    }
    if(mcp.digitalRead(INPUT4_BTN) == 0){
       btn_val_update |= 0b00010000; //Rボタンを1にする   
    }   
    
    // ボタン表示変更
    if((input_val & 0b0001) != (input_val_old & 0b0001)){
      if((input_val & 0b0001) == 0) M5.Lcd.drawJpgFile(SD, "/taiko_don_on.jpg", 0, 180);
      else M5.Lcd.drawJpgFile(SD, "/taiko_don.jpg", 0, 180);
    }
    if((input_val & 0b0010)  != (input_val_old & 0b0010)){
      if((input_val & 0b0010) == 0) M5.Lcd.drawJpgFile(SD, "/taiko_don_on.jpg", 80, 180);
      else M5.Lcd.drawJpgFile(SD, "/taiko_don.jpg", 80, 180);
    }
    if((input_val & 0b0100) != (input_val_old & 0b0100)){
      if((input_val & 0b0100) == 0) M5.Lcd.drawJpgFile(SD, "/taiko_kaxtsu_on.jpg", 160, 180);
      else M5.Lcd.drawJpgFile(SD, "/taiko_kaxtsu.jpg", 160, 180);
    }
    if((input_val & 0b1000) != (input_val_old & 0b1000)){
      if((input_val & 0b1000) == 0) M5.Lcd.drawJpgFile(SD, "/taiko_kaxtsu_on.jpg", 240, 180);
      else M5.Lcd.drawJpgFile(SD, "/taiko_kaxtsu.jpg", 240, 180);
    }
  }
  
  if(jyuji_val_update == 0 && btn_val_update == 0){
    g_btnoff_time = millis();
  }
  
  ///////////////////////////
  // 連打判定
  // (太鼓の達人のみの特殊モード)
  ///////////////////////////
  // RENDA_JUDGE_TIME 以上の時間ボタンが押されている場合、連打モードへ移行
  if(game_idx == 2){
    if(Time_Measure( g_btnoff_time ) > RENDA_JUDGE_TIME){
      if((Time_Measure( g_btnoff_time ) % RENDA_KANKAKU_MS) < (RENDA_KANKAKU_MS/2)){
         jyuji_val_update = 0;
         btn_val_update = 0;
      }
    }
  }

   ////////////////////////////
   // コントローラボタン制御
   ///////////////////////////
   // 上下ボタン
   if((jyuji_val_update & 0b1000) >> 3){ //上の場合
      mcp.pinMode(L_V_JOY, OUTPUT);
      mcp.digitalWrite(L_V_JOY, 0); 
      Serial.println("push UP");
   }else if((jyuji_val_update & 0b0100) >> 2){ //下の場合
      mcp.pinMode(L_V_JOY, OUTPUT);
      mcp.digitalWrite(L_V_JOY, 1); 
      Serial.println("push DOWN");  
   }else{
      mcp.pinMode(L_V_JOY, INPUT);
   }

    // 左右ボタン
   if((jyuji_val_update & 0b0010) >> 1){ //左の場合
      mcp.pinMode(L_H_JOY, OUTPUT);
      mcp.digitalWrite(L_H_JOY, 0);
      Serial.println("push LEFT");
   }else if(jyuji_val_update & 0b0001){  //右の場合
      mcp.pinMode(L_H_JOY, OUTPUT);
      mcp.digitalWrite(L_H_JOY, 1);
      Serial.println("push RIGHT");
   }else{
      mcp.pinMode(L_H_JOY, INPUT);
   }

    // X(青)ボタン
   if((btn_val_update & 0b00000100) >> 2){
      mcp.pinMode(X_BTN, OUTPUT);
      mcp.digitalWrite(X_BTN, 0);
      Serial.println("push X");
   }else{
      mcp.pinMode(X_BTN, INPUT);
   }

   // B(黄)ボタン
   if((btn_val_update & 0b00000010) >> 1){
      mcp.pinMode(B_BTN, OUTPUT);
      mcp.digitalWrite(B_BTN, 0);
      Serial.println("push B");
   }else{
      mcp.pinMode(B_BTN, INPUT);
   }

   // Y(緑)ボタン
   if((btn_val_update & 0b00001000) >> 3){
      mcp.pinMode(Y_BTN, OUTPUT);
      mcp.digitalWrite(Y_BTN, 0);
      Serial.println("push Y");
   }else{
      mcp.pinMode(Y_BTN, INPUT);
   }

   // A(赤)ボタン
   if(btn_val_update & 0b00000001){
      mcp.pinMode(A_BTN, OUTPUT);
      mcp.digitalWrite(A_BTN, 0);
      Serial.println("push A");
//   }else if(digitalRead(INPUT1_BTN) == 0){
//      pinMode(A_BTN, OUTPUT);
//      digitalWrite(A_BTN, 0);
//      Serial.println("push A (Outside)");
//   
   }else{
      mcp.pinMode(A_BTN, INPUT);
   }
    // LZボタン
   if((btn_val_update & 0b10000000) >> 7){
      mcp.pinMode(LZ_BTN, OUTPUT);
      mcp.digitalWrite(LZ_BTN, 0);
      Serial.println("push LZ");
   }else{
      mcp.pinMode(LZ_BTN, INPUT);
   }

   // Lボタン
   if((btn_val_update & 0b01000000) >> 6){
      mcp.pinMode(L_BTN, OUTPUT);
      mcp.digitalWrite(L_BTN, 0);
      Serial.println("push L");
   }else{
      mcp.pinMode(L_BTN, INPUT);
   }

    // RZボタン
   if((btn_val_update & 0b00100000) >> 5){
      mcp.pinMode(RZ_BTN, OUTPUT);
      mcp.digitalWrite(RZ_BTN, 0);
      Serial.println("push RZ");
   }else{
      mcp.pinMode(RZ_BTN, INPUT);
   }

   // Rボタン
   if((btn_val_update & 0b00010000) >> 4){
      mcp.pinMode(R_BTN, OUTPUT);
      mcp.digitalWrite(R_BTN, 0);
      Serial.println("push R");
   }else{
      mcp.pinMode(R_BTN, INPUT);
   }

  // データ更新
  input_val_old = input_val;
  delay(20);
}

uint32_t Time_Measure( uint32_t st_time ){
  return millis() - st_time;
}
