#include <M5Stack.h>
#include <NeoPixelBrightnessBus.h> // instead of NeoPixelBus.h
#include <SparkFunMPU9250-DMP.h>
#include <DFRobotDFPlayerMini.h>

#define BUTTON_PIN 2    // スイッチ入力用GPIO
#define PIXEL_PIN  5    // LED出力用GPIO

///////////////
// LED設定
///////////////
#define PIXEL_COUNT 144     //LED点灯個数
NeoPixelBrightnessBus<NeoGrbFeature, Neo800KbpsMethod> strip(PIXEL_COUNT, PIXEL_PIN);

////////////////////
// DFPlayerMini設定
////////////////////
HardwareSerial DFplayerSerial(2);
DFRobotDFPlayerMini myDFPlayer;
// スピーカ音声ファイル名
#define SE_KIDOU     1
#define SE_RYUSEI    2
#define SE_RAINBOW   3
#define SE_CHEER      4
#define SE_KIRAKIRA  5

//////////////////////
// 加速度センサ設定
//////////////////////
#define ACCVAL_ledmode_TH  150
#define ACCVAL_LEDOFF_TH    20
#define ACCEL_THR_MAX = 700;
#define ACCEL_THR_MIN = 300;
MPU9250_DMP IMU;
uint32_t acc_timer;

/////////////////
// GLOABAL変数
////////////////
uint32_t g_start_time = 0;
int ledmode, ledmode_old=0;
int stop_ledmode;
int play_mode;  // 0: OFF, 1: Loop, 2:Single, 3: Single(ACC OFF)
bool ledmode_is_changed = false;
int g_meteo_cnt = 0;  //meteorRain表示時の積算LED表示位置(積算値)用のグローバル変数



void setup() {
    M5.begin();
    dacWrite(25, 0); // Speaker OFF
    Serial.begin(115200);

    //加速度センサ設定
    Wire.begin();
    if (IMU.begin() != INV_SUCCESS){
      while (1){
         Serial.println("Unable to communicate with MPU-9250");
         Serial.println("Check connections, and try again.");
         Serial.println();
         delay(5000);
       }
    }

    IMU.dmpBegin(DMP_FEATURE_6X_LP_QUAT | // Enable 6-axis quat
               DMP_FEATURE_GYRO_CAL, // Use gyro calibration
              10); // Set DMP FIFO rate to 10 Hz

    //LED設定
    Serial.println("start_led");
    strip.Begin();
    strip.SetBrightness(90);    //LED
    strip.Show();
   
    play_mode = 1;
    stop_ledmode = 0;
    ledmode = -1;
    ledmode_old = -2;
    g_start_time = millis();  //リピートモード用の時間リセット

    // LED表示タスク設定
    TaskHandle_t th;   //マルチタスクハンドル定義
    xTaskCreatePinnedToCore(DisplayLED, "DisplayLED", 4096, NULL, 10, &th, 0); //マルチタスク起動

    acc_timer = micros();
    delay(1000);
    DisplayInitial();

    // DFPlayerMini設定
    DFplayerSerial.begin(9600);
    Serial.println("start_speaker");
    myDFPlayer.begin(DFplayerSerial);
    myDFPlayer.volume(20);  //Set volume value. From 0 to 30
    myDFPlayer.play(SE_KIDOU);  //1:起動完了の音
    delay(2000);
}



void loop() {
    M5.update();

    // Aボタンを押した時、playmode設定
    if(M5.BtnA.wasPressed()){
       Serial.print("change play_mode : ");
       play_mode = (play_mode+1)%4;
       Serial.println(play_mode);
       ledmode_is_changed = true;
       g_start_time = millis();  //リピートモード用時間リセット 
    }  
    
    //真ん中ボタンを押した場合 or 60秒経過時、LEDモードを変更
    if(M5.BtnB.wasPressed()||(play_mode == 1 && Time_Measure(g_start_time) > 60*1000)){
       Serial.print("change led_mode : ");
       stop_ledmode = (stop_ledmode+1)%4;
       if(ledmode < 10) ledmode = stop_ledmode;
       else if(ledmode < 10) ledmode = stop_ledmode+10;
       Serial.println(ledmode);
       g_start_time = millis();  //リピートモード用時間リセット 
       g_meteo_cnt = 0; // 流星イルミネーションカウンタをリセット
    }

    //////////////////////
    // 加速度センサー値取得
    //////////////////////
    double velX, velY, velZ = 0;
    double comVelX, comVelY, comVelZ = 0;
    double comAccX,comAccY,comAccZ=0;
    double gravityX, gravityY, gravityZ=0;
    const double alpha = 0.8;//lowpassfilter

    IMU.update(UPDATE_ACCEL | UPDATE_GYRO | UPDATE_COMPASS);
    double dt = (double)(micros() - acc_timer) / 1000000; // Calculate delta time
    acc_timer = micros();

    // 重力加速度を求める
    gravityX = alpha * gravityX + (1 - alpha) * IMU.calcAccel(IMU.ax);
    gravityY = alpha * gravityY + (1 - alpha) * IMU.calcAccel(IMU.ay);
    gravityZ = alpha * gravityZ + (1 - alpha) * IMU.calcAccel(IMU.az);

    // 補正した加速度
    comAccX = IMU.calcAccel(IMU.ax) - gravityX;
    comAccY = IMU.calcAccel(IMU.ay) - gravityY;
    comAccZ = IMU.calcAccel(IMU.az) - gravityZ;

    velX = velX + comAccX*dt;
    velY = velY + comAccY*dt;
    velZ = velZ + comAccZ*dt;

    Serial.print(comAccX); Serial.print("\t");
    Serial.print(comAccY); Serial.print("\t");
    Serial.print(comAccZ); Serial.println("\t");

    // ある一定値以上の加速度が発生している場合、LED点灯へ
    int ledmode_tmp = stop_ledmode;
    if(comAccX > 0.05 || comAccX < -0.05 || comAccY > 0.05 || comAccY < -0.05){
         ledmode_tmp = stop_ledmode + 10;
    }

     // ボタンCが押されている or 外部スイッチが押されている場合、LED点灯に
    if(M5.BtnC.isPressed() == 1 || digitalRead(BUTTON_PIN) == 0)  ledmode_tmp = stop_ledmode + 10;
  
    if(play_mode != 1) g_start_time = millis();  //リピートモード以外では時間リセット
    
    //////////////////////////
    // M5Stack LCD表示
    /////////////////////////
    M5.Lcd.setCursor(140, 80);
    if(play_mode == 0){
      M5.Lcd.fillRect(140,80, 120,25,(255<<11 | 255<<5 | 255));
      M5.Lcd.setTextSize(2);
      M5.Lcd.setTextColor(BLACK); 
      M5.Lcd.print("LED OFF");
    }else if(play_mode == 2){
      M5.Lcd.fillRect(140,80, 120,25,(255<<11 | 255<<5 | 255));
      M5.Lcd.setTextSize(2);
      M5.Lcd.setTextColor(BLUE); 
      M5.Lcd.print("Single");      
    }else if(play_mode == 3){
      M5.Lcd.fillRect(140,80, 120,25,(255<<11 | 255<<5 | 255));
      M5.Lcd.setTextSize(2);
      M5.Lcd.setTextColor(ORANGE); 
      M5.Lcd.print("ACC OFF");      
    }else if(play_mode == 1){    
      M5.Lcd.fillRect(140,80, 120,25,(255<<11 | 255<<5 | 255));
      M5.Lcd.setTextSize(2);
      M5.Lcd.setTextColor(RED); 
      M5.Lcd.print("Loop  ");       
    }
    if(stop_ledmode == 0 && ledmode < 10){
      M5.Lcd.fillTriangle(70, 105, 70, 135, 90, 120, (0<<11 | 255<<5 | 255));   //矢印表示
    }else if(stop_ledmode == 0 && ledmode >= 10){
      M5.Lcd.fillTriangle(70, 105, 70, 135, 90, 120, (255<<11 | 128<<5 | 0));   //矢印表示
    }else{    
      M5.Lcd.fillTriangle(70, 105, 70, 135, 90, 120, (255<<11 | 255<<5 | 255));  //矢印削除
    }
    if(stop_ledmode == 1 && ledmode < 10){
      M5.Lcd.fillTriangle(70, 135, 70, 165, 90, 150, (0<<11 | 255<<5 | 255));   //矢印表示
    }else if(stop_ledmode == 1 && ledmode >= 10){
      M5.Lcd.fillTriangle(70, 135, 70, 165, 90, 150, (255<<11 | 128<<5 | 0));   //矢印表示
    }else{    
      M5.Lcd.fillTriangle(70, 135, 70, 165, 90, 150, (255<<11 | 255<<5 | 255));  //矢印削除
    }
    if(stop_ledmode == 2 && ledmode < 10){
      M5.Lcd.fillTriangle(70, 165, 70, 195, 90, 180, (0<<11 | 255<<5 | 255));   //矢印表示
    }else if(stop_ledmode == 2 && ledmode >= 10){
      M5.Lcd.fillTriangle(70, 165, 70, 195, 90, 180, (255<<11 | 128<<5 | 0));   //矢印表示
    }else{    
      M5.Lcd.fillTriangle(70, 165, 70, 195, 90, 180, (255<<11 | 255<<5 | 255));  //矢印削除
    }
    if(stop_ledmode == 3 && ledmode < 10){
      M5.Lcd.fillTriangle(70, 195, 70, 225, 90, 210, (0<<11 | 255<<5 | 255));   //矢印表示
    }else if(stop_ledmode == 3 && ledmode >= 10){
      M5.Lcd.fillTriangle(70, 195, 70, 225, 90, 210, (255<<11 | 128<<5 | 0));   //矢印表示
    }else{    
      M5.Lcd.fillTriangle(70, 195, 70, 225, 90, 210, (255<<11 | 255<<5 | 255));  //矢印削除
    }

    //////////////////////////
    // スピーカ再生設定
    /////////////////////////
    if(ledmode != ledmode_old){
       // 再生音選択1
       switch(ledmode){
        case 0:
        case 1:
        case 2:
        case 3:
           myDFPlayer.pause();
           break;              
        case 10:  //流星音
           myDFPlayer.play(SE_RYUSEI);
           break;        
        case 11:  //歓声音
           myDFPlayer.play(SE_CHEER);
           break;
        case 12: //キラキラ
             myDFPlayer.play(SE_KIRAKIRA);
             break;
        case 13:  //レインボー
           myDFPlayer.play(SE_RAINBOW);
           break;
       }
       
       ledmode_is_changed = true;
       Serial.println("ledmode Changed!!");      
    }     


    Serial.println(ledmode);
    ledmode_old = ledmode;
    
    delay(50);
    
}



/////////////////////////////
// LED表示タスク処理
/////////////////////////////
void DisplayLED(void *pvParameters) {
    while(1){
      if(play_mode == 0){
        setAll(RgbColor(0, 0, 0));
        strip.Show(); 
      }else{
         switch(ledmode){
          case 0:   //消灯
            setAll(RgbColor(0, 0, 0));
            break;
          case 10:  //センターに向かって流星表示(赤)
            meteorRainCenter(RgbColor(0xff, 0xff, 0xff),RgbColor(0x3f, 0xff, 0xff), 8, true, 50, true);
            break;
          case 11: //端から端へ流星が移動（緑）
            setAll(RgbColor(0, 0, 0));
            meteorRain(RgbColor(0xff, 0xff, 0xff),RgbColor(0xff, 0x3f, 0xff), 2, true, 20, true);
            break; 
          case 12: //白青ピカピカ
            Sparkle(0xff, 0xff, 0xff, 6, 10);
            Sparkle(0x0,  0x0, 0xff, 6, 10);      
            break;
          case 13:  //レインボー
            rainbowMove(5);   
            break;
         }
      }
      ledmode_is_changed = false;
      delay(50);
    }
}


/////////////////////////////
// LCD初期表示設定
/////////////////////////////
void DisplayInitial(){
    M5.Lcd.fillScreen(WHITE); 
    M5.Lcd.drawJpgFile(SD, "/menu.jpg");
    M5.Lcd.setTextColor(BLACK); 
    M5.Lcd.setTextSize(2);
    M5.Lcd.setCursor(20, 80);
    M5.Lcd.print("PLAY MODE:");
    M5.Lcd.setTextSize(3);
    M5.Lcd.setCursor(100, 110);
    M5.Lcd.print("MODE 1");
    M5.Lcd.setCursor(100, 140);
    M5.Lcd.print("MODE 2");
    M5.Lcd.setCursor(100, 170);
    M5.Lcd.print("MODE 3");
    M5.Lcd.setCursor(100, 200);
    M5.Lcd.print("MODE 4");

    M5.Lcd.setTextSize(1);
    M5.Lcd.setCursor(15, 230);
    M5.Lcd.print("change PLAYMODE");
    M5.Lcd.setCursor(120, 230);
    M5.Lcd.print("change LED");
    M5.Lcd.setCursor(235, 230);
    M5.Lcd.print("fever!!!");
}



void Sparkle(byte red, byte green, byte blue, int count, int SpeedDelay) {
  int i;
  int Pixel[count];
  for(i=0; i<count; i++){
    Pixel[i] = random(PIXEL_COUNT);
  }
  
  for(i=0; i<count; i++){
    //strip_l.SetPixelColor(Pixel[i], RgbColor(red, green, blue));
    strip.SetPixelColor(Pixel[i], RgbColor(red, green, blue));
  }
  //strip_l.Show();
  strip.Show();
  delay(SpeedDelay);
  
  for(i=0; i<count; i++){
    //strip_l.SetPixelColor(Pixel[i], RgbColor( 0, 0, 0));
    strip.SetPixelColor(Pixel[i], RgbColor( 0, 0, 0));
  }
  //strip_l.Show();
  strip.Show();
}



void meteorRain(RgbColor meteor, RgbColor fade, byte meteorSize, boolean meteorRandomDecay, int SpeedDelay, bool reverse) {  
 
  while(!ledmode_is_changed){
    for(int j=0; j<PIXEL_COUNT; j++) {
      if( (!meteorRandomDecay) || (random(10)>5) ) {
        if(reverse == true){       
          fadeToBlack(j, fade.R, fade.G, fade.B); 
        }else{
          fadeToBlack((PIXEL_COUNT-1-j), fade.R, fade.G, fade.B);
        }
      }
    }
    
    for(int j = 0; j < meteorSize; j++) {
      if( ( g_meteo_cnt-j <PIXEL_COUNT) && (g_meteo_cnt-j>=0) ) {
         if(reverse == true){ 
            strip.SetPixelColor(g_meteo_cnt-j, meteor);
         }else{
            strip.SetPixelColor(PIXEL_COUNT-1-(g_meteo_cnt-j), meteor);    
         }
      } 
    }
   
    strip.Show();

    if(g_meteo_cnt >= PIXEL_COUNT+PIXEL_COUNT/4){
      g_meteo_cnt = 0;
    }
    if(ledmode_is_changed == true) return;
    
    g_meteo_cnt++;
    delay(SpeedDelay);
  }
}


void meteorRainCenter(RgbColor meteor, RgbColor fade, byte meteorSize, boolean meteorRandomDecay, int SpeedDelay, bool reverse) {  
 
  for(int i = 0; i < PIXEL_COUNT/2*1.5; i++) { 
    for(int j=0; j<PIXEL_COUNT/2; j++) {
      if( (!meteorRandomDecay) || (random(10)>5) ) {
        if(reverse == true){       
          fadeToBlack(j, fade.R, fade.G, fade.B);
          fadeToBlack(PIXEL_COUNT-1-j, fade.R, fade.G, fade.B); 
        }else{
          fadeToBlack((PIXEL_COUNT/2-1-j), fade.R, fade.G, fade.B);
          fadeToBlack((PIXEL_COUNT/2-1+j), fade.R, fade.G, fade.B);
        }
      }
    }
    
    // draw meteor
    for(int j = 0; j < meteorSize; j++) {
      if( ( i-j < PIXEL_COUNT/2) && (i-j>=0) ) {  
         if(reverse == true){ 
            strip.SetPixelColor(i-j, meteor);
            strip.SetPixelColor(PIXEL_COUNT-(i-j), meteor);
         }else{  
            strip.SetPixelColor(PIXEL_COUNT/2-1-(i-j), meteor);
            strip.SetPixelColor(PIXEL_COUNT/2+1+(i-j), meteor);
         }
      } 
    }
    strip.Show();

    if(ledmode_is_changed == true) return;

    delay(SpeedDelay);
  }
  return;
}

void fadeToBlack(int ledNo, byte fade_r, byte fade_g, byte fade_b) {
    // NeoPixel
    RgbColor oldColor;
    uint8_t r, g, b;
    int value;
    
    oldColor = strip.GetPixelColor(ledNo);

    r=(oldColor.R<=10)? 0 : (int)oldColor.R -(oldColor.R * fade_r/256);
    g=(oldColor.G<=10)? 0 : (int)oldColor.G -(oldColor.G * fade_g/256);
    b=(oldColor.B<=10)? 0 : (int)oldColor.B -(oldColor.B * fade_b/256);
    
    strip.SetPixelColor(ledNo, RgbColor(r,g,b));
}






void rainbowMove(int interval_ms) {
  uint16_t i, j;
 

  for(j=0; j<256; j++) { // 5 cycles of all colors on wheel
    for (int i=0; i< PIXEL_COUNT; i++) {
        strip.SetPixelColor(i, rainbowColor(((i * 256 / PIXEL_COUNT) + j) & 255));
    }
    strip.Show();
    if(ledmode_is_changed == true) return;
    delay(interval_ms);
  }
}


RgbColor rainbowColor(byte color_num) {
  color_num = 255 - color_num;
  if(color_num < 85) {
    return RgbColor(255 - color_num * 3, 0, color_num * 3);
  }
  if(color_num < 170) {
    color_num -= 85;
    return RgbColor(0, color_num * 3, 255 - color_num * 3);
  }
  color_num -= 170;
  return RgbColor(color_num * 3, 255 - color_num * 3, 0);
}




void setAll(RgbColor color) {
  for(int i = 0; i <PIXEL_COUNT; i++ ) {
    strip.SetPixelColor(i, color); 
  }
  strip.Show();
}

uint32_t Time_Measure( uint32_t st_time ){
  return millis() - st_time;
}
