#include <M5Atom.h>
#include <WiFi.h>
#include <WiFiUdp.h>

#define F_PWM_PIN 23
#define R_PWM_PIN 33

#define PULSE_START  60  //駆動に最低限必要な最低PCMパルス数
#define PULSE_STEP   15  //step毎に加算していくPCMパルス数 (9段階)

const char ssid[] = "m5mobility"; // SSID
const char password[] = "abcde";  // password
const int WifiPort = 8000;      // ポート番号

const IPAddress HostIP(192, 168, 4, 1);       // IPアドレス
const IPAddress ClientIP(192, 168, 4, 2);       // IPアドレス
const IPAddress subnet(255, 255, 255, 0); // サブネットマスク
const IPAddress gateway(192,168, 4, 0);
const IPAddress dns(192, 168, 4, 0);
WiFiUDP Udp;
bool wifi_connect;
#define CMD_SIZE  5

uint32_t  g_wifi_interval_time = 0;
uint8_t WiFibuff[CMD_SIZE];
int vel = 0;		//プラレール速度値 (0-9の9段階)

/////////////////////////////
//m5atom用LEDマトリックス設定
///////////////////////////// 
float r, g, b;
float ratio_h = 2.5;
float ratio_l = 0.6;  
uint8_t DisBuff[2 + 5 * 5 * 3];

void setBuffP(uint8_t posData, uint8_t Rdata, uint8_t Gdata, uint8_t Bdata){
    DisBuff[2 + posData * 3 + 0] = Rdata;
    DisBuff[2 + posData * 3 + 1] = Gdata;
    DisBuff[2 + posData * 3 + 2] = Bdata;
}
void setBuff(uint8_t Rdata, uint8_t Gdata, uint8_t Bdata){
    for (uint8_t i = 0; i < 25; i++)
        setBuffP(i, Rdata, Gdata, Bdata);
}

void shftBuff(){
    for (uint8_t i = 24; i > 0; i--)    {
        for (uint8_t j = 0; j < 3; j++)
            DisBuff[2 + i * 3 + j] = DisBuff[2 + (i-1) * 3 + j];
    }
}

void setup() 
{
  //PIN設定
  pinMode(F_PWM_PIN, OUTPUT);
  pinMode(R_PWM_PIN, OUTPUT);
  digitalWrite(F_PWM_PIN, 0);
  digitalWrite(R_PWM_PIN, 0);

  //PWM設定
  ledcSetup(0, 490, 8);
  ledcSetup(1, 490, 8);

  ledcAttachPin(F_PWM_PIN, 0);  //Ch0: モーター(正回転側)
  ledcAttachPin(R_PWM_PIN, 1);  //CH1: モーター(逆回転側)

  ledcWrite(0, 0);
  ledcWrite(1, 0);

  DisBuff[0] = 0x05;
  DisBuff[1] = 0x05;
  M5.begin(false, false, true);
  digitalWrite(F_PWM_PIN, 0);
  digitalWrite(R_PWM_PIN, 0);
  
  setBuff(0xf0, 0xf0, 0x00);
  M5.dis.displaybuff(DisBuff);
  
  Serial.begin(115200);
  motor_inst = 0;
  vel = 0;
  emergency_mode = 0;
  
  // Wifi設定 (SSID&パス設定)
  wifi_connect = false;
  WiFi.mode(WIFI_AP);
  WiFi.softAP(ssid, password);
  WiFi.softAPConfig(HostIP, HostIP, subnet); // IPアドレス設定

  Serial.print("AP IP address: ");
  IPAddress myAddress = WiFi.softAPIP();
  Serial.println(myAddress);

  Udp.begin(WifiPort);  // UDP通信開始
  Serial.println("Starting UDP");
  
  Serial.print("Local port: ");
  Serial.println(WifiPort);

  delay(3000);
}

void loop() {

loop_start:

  int packet_size = Udp.parsePacket();

  if (packet_size > 0) {  
    if(wifi_connect == false){
       wifi_connect = true;
    }
    g_wifi_interval_time = millis();  //無線インターバル計測タイマーリセット 
    Udp.read(WiFibuff, CMD_SIZE);
    Udp.flush();    
    
    String s_buf = String((char*)WiFibuff);
    if(s_buf.substring(1,4) == "VEL"){
      vel = (s_buf.substring(4,5)).toInt();
    }
  }

  // 無線通信状態確認
  if(Time_Measure(g_wifi_interval_time) > 500){
     wifi_connect = false;
     vel = 0;
     Serial.println("Wifi Discconnect");
  }

  // Motor制御
  // 左右モーターの速度計算
  if(vel > 0){
    ledcWrite(0,abs(calc_MotorPWM(vel,  PULSE_START)));
    ledcWrite(1, 0);
  }else{
    ledcWrite(0, 255);
    ledcWrite(1, 255);
  }


  // Display表示
  //先頭表示
  int i;
  if(emergency_mode == 1){
      setBuff(0xff, 0x00, 0x00);
      M5.dis.displaybuff(DisBuff);
  }else if(vel < 0){
    M5.dis.drawpix(10, 0x80ff00);
    M5.dis.drawpix(11, 0x80ff00);
    M5.dis.drawpix(12, 0x80ff00);
    M5.dis.drawpix(13, 0x80ff00);
    M5.dis.drawpix(14, 0x80ff00);
    M5.dis.drawpix(16, 0x80ff00);
    M5.dis.drawpix(17, 0x80ff00);
    M5.dis.drawpix(18, 0x80ff00);
    M5.dis.drawpix(22, 0x80ff00);
  }else{
    for(i=0; i<5; i++){
      if(vel >= 9) M5.dis.drawpix(i, 0x8f007f);  //blue 0x0000ff
      else                M5.dis.drawpix(i, 0x000000);
    }
    for(i=5; i<10; i++){
      if(vel >= 7) M5.dis.drawpix(i, 0x6f009f);  //blue 0x0000ff
      else                M5.dis.drawpix(i, 0x000000);
   }
   for(i=10; i<15; i++){
     if(vel >= 5) M5.dis.drawpix(i, 0x4f00bf);  //blue 0x0000ff
     else                M5.dis.drawpix(i, 0x000000);
   }
    for(i=15; i<20; i++){
      if(vel >= 3) M5.dis.drawpix(i, 0x2f00df);  //blue 0x0000ff
      else                M5.dis.drawpix(i, 0x000000);
    }
    for(i=20; i<25; i++){
      if(vel >= 1) M5.dis.drawpix(i, 0x0000ff);  //blue 0x0000ff
      else if(i==22 && wifi_connect == true) M5.dis.drawpix(22, 0xf0f000);
      else     M5.dis.drawpix(i, 0x000000);
    }
  }

  
  delay(100);
}

void rcvWiFi() {

}


int calc_MotorPWM(int vel, int pulse_start){
  int pwm;
  if(vel > 0){
    pwm =  (int)(pulse_start + vel * PULSE_STEP);  
  }else if(vel < 0){
    pwm =  (int)(-1 * pulse_start + vel * PULSE_STEP);     
  }else{
    pwm = 0;
  }

  //限度設定
  if(pwm >  250) pwm =  250;
  if(pwm < -250) pwm = -250;
  
  return pwm;
}


uint32_t Time_Measure( uint32_t st_time ){
  return millis() - st_time;
}
