#include <M5Stack.h>
#include <Wire.h>
#include <Adafruit_MCP23X17.h>
#include <NeoPixelBrightnessBus.h> // instead of NeoPixelBus.h
#include <WiFi.h>
#include <WiFiUdp.h>
#include <DFRobotDFPlayerMini.h>

const char ssid[] = "m5mobility"; // SSID
const char password[] = "abcde";  // password
const int WifiPort = 8000;      // ポート番号

const IPAddress HostIP(192, 168, 4, 1);       // IPアドレス
const IPAddress ClientIP(192, 168, 4, 2);       // IPアドレス
const IPAddress subnet(255, 255, 255, 0); // サブネットマスク
const IPAddress gateway(192,168, 4, 0);
const IPAddress dns(192, 168, 4, 0);
WiFiUDP Udp;
bool wifi_connect;
#define CMD_SIZE  5

//MCP23017 (GPIOエキスパンダ)の使い方
// http://project59.blog.fc2.com/blog-entry-441.html
// ポイント : ResetはVcc、アドレスはGNDに接続する事
#define ACCEL_A0_PIN    8
#define ACCEL_A1_PIN    9
#define ACCEL_A2_PIN   10
#define BRAKE_A0_PIN    0
#define BRAKE_A1_PIN    1
#define BRAKE_A2_PIN    2
#define BRAKE_A3_PIN    3
#define BUTTON_PIN      4

////////////////////
// NeoPixel LEDピン
///////////////////
#define PIXEL_PIN 5
#define PIXEL_COUNT 9
NeoPixelBrightnessBus<NeoGrbFeature, Neo800KbpsMethod> strip(PIXEL_COUNT, PIXEL_PIN);

//////////////////
// DFplayerMini
/////////////////

HardwareSerial DFplayerSerial(2);
DFRobotDFPlayerMini myDFPlayer;
#define DFP_TRAIN_KIDOU 1
#define DFP_TRAIN_BELL  2
#define DFP_TRAIN_DOOR  3
#define DFP_TRAIN_START 4
#define DFP_TRAIN_DRIVE 5
#define DFP_TRAIN_STOP  6
#define DFP_TRAIN_BRAKE 7
#define DFP_TRAIN_HORN  8
#define DFP_TRAIN_LEFT  9
#define DFP_TRAIN_RIGHT 10
#define DFP_TRAIN_BACK  11

#define DFP_VOLUME 20

//GLOBAL変数
int g_accel, g_accel_old = 0;
int g_brake, g_brake_old = 0;
int g_direct,g_direct_old = 0;
int g_button, g_button_old = 0;
int g_stick_pos = 0;    // 0: センター、4: 上、-4：下 に加えて、[+1:右、-1:左]を補正
int g_stick_pos_now = 15;
bool g_stick_pos_is_changed = false;
uint32_t g_disp_time = 0;
float g_vel_raw;
int g_vel_set, g_vel_set_old = 0; 
int g_sound_cnt = 0;

int dispsts, dispsts_old = 0;  // 0:通常、1:発車します、2:停止、3:緊急停止

Adafruit_MCP23X17 mcp;

void setup()
{
  M5.begin();
  Serial.begin(115200);
  M5.Lcd.fillScreen(BLACK);

  //LCD画面設定
  M5.Lcd.fillScreen(WHITE); 
  M5.Lcd.drawJpgFile(SD, "/menu.jpg");

  mcp.begin_I2C();

  mcp.pinMode(ACCEL_A0_PIN, INPUT_PULLUP);
  mcp.pinMode(ACCEL_A1_PIN, INPUT_PULLUP);
  mcp.pinMode(ACCEL_A2_PIN, INPUT_PULLUP);
  mcp.pinMode(BRAKE_A0_PIN, INPUT_PULLUP);
  mcp.pinMode(BRAKE_A1_PIN, INPUT_PULLUP);
  mcp.pinMode(BRAKE_A2_PIN, INPUT_PULLUP);
  mcp.pinMode(BRAKE_A3_PIN, INPUT_PULLUP);
  mcp.pinMode(BUTTON_PIN, INPUT_PULLUP);
  
  // Wifi設定 (SSID&パス設定)  
  WiFi.config(ClientIP, gateway, subnet, dns);
  WiFi.begin(ssid, password);
  Udp.begin(WifiPort);
     
  Serial.println("start_connect");
  int i=0;
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
    if(i >= 10){
      break;
    }
    i++;
  }
  
  if( i >= 10){
    Serial.println("Wifi Disconnected...");
  }else{
    Serial.println("");
    Serial.println("WiFi connected.");
    Serial.println("IP address: ");
    Serial.println(WiFi.localIP());
  }

  //LED設定
  pinMode(PIXEL_PIN, OUTPUT);
  strip.Begin();
  strip.SetBrightness(150);
  strip.Show();
  rainbowMove(2);
  setAll(RgbColor(0, 0, 0));
  
  // Speaker設定
  DFplayerSerial.begin(9600);
  Serial.println("start_speaker");
  myDFPlayer.begin(DFplayerSerial);
  myDFPlayer.volume(DFP_VOLUME);  //Set volume value. From 0 to 30
  // スピーカ再生
  myDFPlayer.play(DFP_TRAIN_KIDOU);  //1:起動完了の音

  g_vel_raw = 0.0;

  delay(2000);

  // LED表示タスク設定
  TaskHandle_t th; //マルチタスクハンドル定義
  xTaskCreatePinnedToCore(DisplayLCD, "DisplayLCD", 4096, NULL, 10, &th, 0); //マルチタスク起動
}


/////////////////////////////
// Main処理
/////////////////////////////
void loop(){
  g_accel =  mcp.digitalRead(ACCEL_A2_PIN)*4 + mcp.digitalRead(ACCEL_A1_PIN)*2 + mcp.digitalRead(ACCEL_A0_PIN);
  g_brake =  mcp.digitalRead(BRAKE_A3_PIN)*8 + mcp.digitalRead(BRAKE_A2_PIN)*4 + mcp.digitalRead(BRAKE_A1_PIN)*2 + mcp.digitalRead(BRAKE_A0_PIN);
  if(g_brake == 0)  g_brake = g_brake_old;
  g_button = mcp.digitalRead(BUTTON_PIN);
    
  if(g_stick_pos != g_stick_pos_now){
     g_stick_pos_is_changed = true;
  }

  
  //速度決定
  if(g_accel != 0 && g_accel != 7 && g_brake != 0){ //アクセル、ブレーキ不定値はスキップ
    if(g_brake >= 9){
      g_vel_raw = 0;      
    }else{
      g_vel_raw += 0.03 * (g_accel-1) - 0.04 * (g_brake-1);
      // アクセルを離した時にゆるやかに減速する
      if((g_accel-1)*1.3 < g_vel_raw) g_vel_raw -= 0.03*(g_vel_raw-1.3*(g_accel-1));
    }
    if(g_vel_raw < 0 )      g_vel_raw = 0;
    else if(g_vel_raw > 9 ) g_vel_raw = 9;

    g_vel_set = (int)g_vel_raw;

    // アクセルONのタイミングだけ早める
    if(g_vel_raw >= 0.5 && g_vel_raw < 1.0) g_vel_set = 1;
  } 

  // UDP通信プロトコル (5byte固定)
  //  #VELx  xは速度(0-9の9段階) 

  byte wifi_buf[CMD_SIZE];
  String  m_buf = "#VEL";
  m_buf.concat(String(g_vel_set));
  m_buf.getBytes(wifi_buf, CMD_SIZE);
  if (Udp.beginPacket(HostIP,  WifiPort)) { 
    Udp.write(wifi_buf,CMD_SIZE);
    Udp.endPacket();
  }

  /////////////////////
  // スピーカ再生ファイル設定
  /////////////////////
  if(g_button == 0 && g_button_old ==1){
    if(g_vel_set > 0){
       myDFPlayer.volume(DFP_VOLUME);
       myDFPlayer.play(DFP_TRAIN_HORN);
    }else{
       if(g_sound_cnt%2 == 0){
         myDFPlayer.volume(DFP_VOLUME);
         myDFPlayer.play(DFP_TRAIN_BELL);
       }else{
         myDFPlayer.volume(DFP_VOLUME);
         myDFPlayer.play(DFP_TRAIN_DOOR);          
       }
       g_sound_cnt++;
    }
  }

  //アクセルをかけた場合は発車音
  if(g_accel >= 2 && g_accel_old ==1){
       myDFPlayer.volume(DFP_VOLUME);
       myDFPlayer.play(DFP_TRAIN_START);
  }
  //アクセルを0にした場合は停車音
  if(g_accel == 1 && g_accel_old >=2){
       myDFPlayer.volume(DFP_VOLUME);
       myDFPlayer.play(DFP_TRAIN_STOP);
  }
  //ブレーキをかけた場合はブレーキ音
  if(g_brake >= 9 && g_brake_old <9){
       myDFPlayer.volume(DFP_VOLUME);
       myDFPlayer.play(DFP_TRAIN_BRAKE);
  }

  /////////////////////
  // LED表示色 設定
  /////////////////////
  // ACCELL量によって色を決める
  int i;
    // 現在の速度値を元に左から順番に青色系で埋める
    for(i=0; i<PIXEL_COUNT; i++){
      if(i < g_vel_set) strip.SetPixelColor(i, RgbColor(0, 25*i, 255-20*i));
      else strip.SetPixelColor(i, RgbColor(0, 0, 0));
    }
    //ブレーキをかけている場合、右から順番に赤色系で埋める
    for(i=0; i<(g_brake-1); i++){
      strip.SetPixelColor(PIXEL_COUNT-i, RgbColor(255-5*i, 0, 10*i));
    }
  //}
  strip.Show();
  
  g_accel_old = g_accel;
  g_brake_old = g_brake;
  g_button_old = g_button;
  delay(100);  //安定待ち

}


/////////////////////////////
// LCD表示タスク処理
/////////////////////////////
void DisplayLCD(void *pvParameters) {
  while(1){
   M5.update();

  //ディスプレイ下部の表示切替
  if(g_vel_set == 0) g_disp_time = millis();
  
  if(g_vel_set == 0)  dispsts = 2;
  else if(g_vel_set != 0 && Time_Measure(g_disp_time) < 3000) dispsts = 1;
  else dispsts = 0;
  
  if(dispsts != dispsts_old){
    M5.Lcd.drawJpgFile(SD, "/train.jpg", 0, 0);
    if(dispsts == 0)  M5.Lcd.drawJpgFile(SD, "/front.jpg", 0, 160);
    if(dispsts == 1)  M5.Lcd.drawJpgFile(SD, "/start.jpg", 0, 160);
    if(dispsts == 2)  M5.Lcd.drawJpgFile(SD, "/stop.jpg", 0, 160);
    if(dispsts == 3)  M5.Lcd.drawJpgFile(SD, "/emergency.jpg", 0, 160);
  }
  
  dispsts_old = dispsts;
  delay(100);
  
    }
}


void rainbowMove(int interval_ms) {
  uint16_t i, j;
 

  for(j=0; j<256; j++) { // 5 cycles of all colors on wheel
    for (int i=0; i< PIXEL_COUNT; i++) {
        strip.SetPixelColor(i, rainbowColor(((i * 256 / PIXEL_COUNT) + j) & 255));
    }
    strip.Show();
    delay(interval_ms);
  }
}


RgbColor rainbowColor(byte color_num) {
  color_num = 255 - color_num;
  if(color_num < 85) {
    return RgbColor(255 - color_num * 3, 0, color_num * 3);
  }
  if(color_num < 170) {
    color_num -= 85;
    return RgbColor(0, color_num * 3, 255 - color_num * 3);
  }
  color_num -= 170;
  return RgbColor(color_num * 3, 255 - color_num * 3, 0);
}


void setAll(RgbColor color) {
  for(int i = 0; i <PIXEL_COUNT; i++ ) {
    strip.SetPixelColor(i, color); 
  }
  strip.Show();
}



uint32_t Time_Measure( uint32_t st_time ){
  return millis() - st_time;
}
